"""Module that implements the VivintPubNubSubscribeListener class."""
import logging
from typing import Callable, Union

import pubnub.models.consumer.common
import pubnub.models.consumer.pubsub
import pubnub.pubnub
import pubnub.pubnub_asyncio
from pubnub.callbacks import SubscribeCallback

PN_SUBSCRIBE_KEY = "sub-c-6fb03d68-6a78-11e2-ae8f-12313f022c90"
PN_CHANNEL = "PlatformChannel"

_LOGGER = logging.getLogger(__name__)


class VivintPubNubSubscribeListener(SubscribeCallback):
    """PubNub Subscribe Listener."""

    def __init__(self, message_received_callback: Callable):
        super().__init__()
        self.__message_received_callback = message_received_callback

    def status(
        self,
        pubnub: Union[pubnub.pubnub.PubNub, pubnub.pubnub_asyncio.PubNubAsyncio],
        status: pubnub.models.consumer.common.PNStatus,
    ) -> None:
        """Handler called on status updates."""
        _LOGGER.info(
            f"pubnub status update {status.category}, error: {status.error}, error_data: {status.error_data}"
        )

    def message(
        self,
        pubnub: Union[pubnub.pubnub.PubNub, pubnub.pubnub_asyncio.PubNubAsyncio],
        message: pubnub.models.consumer.pubsub.PNMessageResult,
    ) -> None:
        """Handler called on each message received."""
        self.__message_received_callback(message.message)



"""The exceptions used by the pyvivint library."""


class VivintError(Exception):
    """General pyvivint exception occurred."""


class VivintSkyApiError(VivintError):
    """VivintSky API related exception occurred."""


class VivintSkyApiAuthenticationError(VivintSkyApiError):
    """VivintSky API authentication related error occurred."""


class VivintSkyApiExpiredCookieError(VivintSkyApiError):
    """VivintSky API cookie expired error occurred."""


class VivintSkyApiMissingCookieError(VivintSkyApiError):
    """VivintSky API missing cookie error occurred."""



"""Module that implements the Vivint class."""
import asyncio
import json
import logging
from typing import List, Optional

import aiohttp

import pubnub
import pyvivint.system
from pubnub.enums import PNHeartbeatNotificationOptions, PNReconnectionPolicy
from pubnub.pnconfiguration import PNConfiguration
from pubnub.pubnub_asyncio import PubNubAsyncio
from pyvivint.constants import AuthUserAttribute
from pyvivint.constants import PubNubMessageAttribute as MessageAttributes
from pyvivint.pubnub import PN_CHANNEL, PN_SUBSCRIBE_KEY, VivintPubNubSubscribeListener
from pyvivint.system import System
from pyvivint.utils import add_async_job, first_or_none
from pyvivint.vivintskyapi import VivintSkyApi

_LOGGER = logging.getLogger(__name__)


class Vivint:
    """Class for interacting with VivintSky API using asyncio"""

    def __init__(
        self,
        username: str,
        password: str,
        client_session: Optional[aiohttp.ClientSession] = None,
    ):
        self.__pubnub: pubnub.pubnub_asyncio.PubNubAsyncio = None
        self.vivintskyapi = VivintSkyApi(username, password, client_session)
        self.systems: List[pyvivint.system.System] = []

    async def connect(
        self,
        load_devices: bool = False,
        subscribe_for_realtime_updates: bool = False,
    ) -> None:
        """Connects to vivintsky cloud service."""

        _LOGGER.debug("connecting to vivintsky")
        # initialize the vivintsky cloud session
        authuser_data = await self.vivintskyapi.connect()

        # load all systems, panels and devices
        if load_devices:
            _LOGGER.debug("loading devices")
            await self.refresh(authuser_data)

        # subscribe to pubnub for realtime updates
        if subscribe_for_realtime_updates:
            _LOGGER.debug("subscribing to pubnub for realtime updates")
            await self.subscribe_for_realtime_updates(authuser_data)

    async def disconnect(self) -> None:
        _LOGGER.debug("disconnecting from vivintsky")
        await self.vivintskyapi.disconnect()

    async def refresh(self, authuser_data: dict = None) -> None:
        # make a call to vivint's authuser endpoint to get a list of all the system_accounts (locations) & panels if not supplied
        if not authuser_data:
            authuser_data = await self.vivintskyapi.get_authuser_data()

        # for each system_account, make another call to load all the devices
        for system_data in authuser_data["u"]["system"]:
            # is this an existing account_system?
            system = first_or_none(
                self.systems, lambda system: system.id == system_data["panid"]
            )
            if system:
                await system.refresh()
            else:
                full_system_data = await self.vivintskyapi.get_system_data(
                    system_data["panid"]
                )
                self.systems.append(
                    System(system_data.get("sn"), full_system_data, self.vivintskyapi)
                )

        _LOGGER.debug(f"loaded {len(self.systems)} system(s)")

    async def subscribe_for_realtime_updates(self, authuser_data: dict = None) -> None:
        """Subscribes to PubNub for realtime updates."""
        # make a call to vivint's authuser endpoint to get message broadcast channel if not supplied
        if not authuser_data:
            authuser_data = await self.vivintskyapi.get_authuser_data()

        pubnub.set_stream_logger("pubnub", logging.INFO)

        pnconfig = PNConfiguration()
        pnconfig.subscribe_key = PN_SUBSCRIBE_KEY
        pnconfig.ssl = True
        pnconfig.reconnect_policy = PNReconnectionPolicy.LINEAR
        pnconfig.heartbeat_notification_options = PNHeartbeatNotificationOptions.ALL

        self.__pubnub = PubNubAsyncio(pnconfig)
        self.__pubnub.add_listener(
            VivintPubNubSubscribeListener(self.handle_pubnub_message)
        )

        pn_channel = f"{PN_CHANNEL}#{authuser_data[AuthUserAttribute.USERS][AuthUserAttribute.UserAttribute.MESSAGE_BROADCAST_CHANNEL]}"
        self.__pubnub.subscribe().channels(pn_channel).with_presence().execute()

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message."""
        _LOGGER.info(f"message: {json.dumps(message)}")

        panel_id = message.get(MessageAttributes.PANEL_ID)
        if not panel_id:
            _LOGGER.info("ignoring pubnub message. No panel_id specified")
            return

        system = first_or_none(self.systems, lambda system: system.id == panel_id)
        if not system:
            _LOGGER.info(f"no system found with id {panel_id}")
            return

        system.handle_pubnub_message(message)



"""Module that defines various constants."""


class Constant:

    __names = {}

    @classmethod
    def name(cls, state):
        if not cls.__names:
            cls.__names = {
                value: key
                for key, value in cls.__dict__.items()
                if isinstance(value, int)
            }

        return cls.__names[state]


class AuthUserAttribute(Constant):
    """AuthUser attributes."""

    ID_TOKEN = "id_token"
    IS_READ_ONLY = "is_read_only"
    KEEP_SIGNED_IN = "keep_signed_in"
    RELAY_SERVER = "rs"
    USERS = "u"

    class UserAttribute(Constant):
        """User Attributes"""

        DOCUMENT_SEQUENCE = "DocumentSequence"
        EMAIL = "e"
        GHOME = "ghome"
        GROUP_IDS = "grpid"
        ID = "_id"
        MESSAGE_BROADCAST_CHANNEL = "mbc"
        NAME = "n"
        PING_ID = "pngid"
        RESTRICTED_SYSTEM = "rsystem"
        SMART_HOME_SYSTEM = "smarthomesystem"
        SETTINGS = "stg"
        SYSTEM = "system"
        TIMESTAMP = "ts"


class PubNubMessageAttribute(Constant):
    """PubNub message attributes."""

    DATA = "da"
    DEVICES = "d"
    OPERATION = "op"
    PANEL_ID = "panid"
    PARTITION_ID = "parid"
    TYPE = "t"


class VivintDeviceAttribute(Constant):
    """Vivint device attributes."""

    BATTERY_LEVEL = "bl"
    CURRENT_SOFTWARE_VERSION = "csv"
    FIRMWARE_VERSION = "fwv"
    ID = "_id"
    LOW_BATTERY = "lb"
    NAME = "n"
    ONLINE = "ol"
    PANEL_ID = "panid"
    SERIAL_NUMBER = "ser"
    SERIAL_NUMBER_32_BIT = "ser32"
    STATE = "s"
    TYPE = "t"


class AlarmPanelAttribute(VivintDeviceAttribute):
    """Alarm panel attributes."""

    DEVICES = "d"
    PARTITION_ID = "parid"


class CameraAttribute(VivintDeviceAttribute):
    """Camera attributes."""

    ACTUAL_TYPE = "act"
    CAMERA_IP_ADDRESS = "caip"
    CAMERA_IP_PORT = "cap"
    CAMERA_MAC = "cmac"
    CAMERA_PRIVACY = "cpri"
    CAMERA_THUMBNAIL_DATE = "ctd"
    CAPTURE_CLIP_ON_MOTION = "ccom"
    SOFTWARE_VERSION = "sv"
    WIRELESS_SIGNAL_STRENGTH = "wiss"


class WirelessSensorAttribute(VivintDeviceAttribute):
    """Wireless sensor attributes."""

    BYPASSED = "b"
    EQUIPMENT_CODE = "ec"
    EQUIPMENT_TYPE = "eqt"
    SENSOR_FIRMWARE_VERSION = "sensor_firmware_version"
    SENSOR_TYPE = "set"


class ZWaveDeviceAttribute(VivintDeviceAttribute):
    """ZWave device attributes."""

    OPERATION_COUNT = "opc"
    OPERATION_FAULT_CODE = "opfc"



"""Module that defines various enums."""
from enum import Enum, IntEnum, unique


@unique
class ArmedState(IntEnum):
    """Armed state."""

    DISARMED = 0
    ARMING_AWAY_IN_EXIT_DELAY = 1
    ARMING_STAY_IN_EXIT_DELAY = 2
    ARMED_STAY = 3
    ARMED_AWAY = 4
    ARMED_STAY_IN_ENTRY_DELAY = 5
    ARMED_AWAY_IN_ENTRY_DELAY = 6
    ALARM = 7
    ALARM_FIRE = 8
    DISABLED = 11
    WALK_TEST = 12


@unique
class DeviceType(Enum):
    """Vivint supported device types."""

    CAMERA = "camera_device"
    DOOR_LOCK = "door_lock_device"
    GARAGE_DOOR = "garage_door_device"
    LIGHT_MODULE = "multilevel_switch_device"
    THERMOSTAT = "thermostat_device"
    TOUCH_PANEL = "primary_touch_link_device"
    WIRELESS_SENSOR = "wireless_sensor"
    UNKNOWN = None

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


@unique
class EquipmentCode(IntEnum):
    """Vivint equipment code."""

    APOLLO_COMBO_CO = 1422
    APOLLO_COMBO_SMOKE = 1322
    CARBON_MONOXIDE_DETECTOR_345_MHZ = 1254
    CO1_CO = 860
    CO1_CO_CANADA = 859
    CO3_2_GIG_CO = 1026
    DBELL1_2_GIG_DOORBELL = 1063
    DW10_THIN_DOOR_WINDOW = 862
    DW11_THIN_DOOR_WINDOW = 1251
    DW20_RECESSED_DOOR = 863
    DW21_R_RECESSED_DOOR = 1252
    EXISTING_CO = 692
    EXISTING_DOOR_WINDOW_CONTACT = 655
    EXISTING_FLOOD_TEMP = 556
    EXISTING_GLASS_BREAK = 475
    EXISTING_HEAT = 708
    EXISTING_KEY_FOB_REMOTE = 577
    EXISTING_MOTION_DETECTOR = 609
    EXISTING_SMOKE = 616
    FIREFIGHTER_AUDIO_DETECTOR = 1269
    GARAGE01_RESOLUTION_TILT = 1061
    GB1_GLASS_BREAK = 864
    GB2_GLASS_BREAK = 1248
    HW_DW_5816 = 673
    HW_FLOOD_SENSOR_5821 = 624
    HW_GLASS_BREAK_5853 = 519
    HW_HEAT_SENSOR_5809 = 557
    HW_PANIC_PENDANT_5802_MN2 = 491
    HW_PIR_5890 = 533
    HW_PIR_5894_PI = 530
    HW_R_DW_5818_MNL = 470
    HW_SMOKE_5808_W3 = 589
    KEY1_345_4_BUTTON_KEY_FOB_REMOTE = 866
    OTHER = 0
    PAD1_345_WIRELESS_KEYPAD = 867
    PANIC1 = 868
    PANIC2 = 1253
    PIR1_MOTION = 869
    PIR2_MOTION = 1249
    RE219_FLOOD_SENSOR = 1128
    RE220_T_2_GIG_REPEATER = 1144
    RE224_DT_DSC_TRANSLATOR = 1208
    RE224_GT_GE_TRANSLATOR = 941
    RE508_X_REPEATER = 2832
    RE524_X_WIRELESS_TAKEOVER = 2830
    REPEATER_345_MHZ = 2081
    SECURE_KEY_345_MHZ = 1250
    SMKE1_SMOKE = 872
    SMKE1_SMOKE_CANADA = 871
    SMKT2_GE_SMOKE_HEAT = 895
    SMKT3_2_GIG = 1058
    SMKT6_2_GIG = 1066
    SWS1_SMART_WATER_SENSOR = 1264
    TAKE_TAKEOVER = 873
    TILT_SENSOR_2_GIG_345 = 2831
    VS_CO3_DETECTOR = 1266
    VS_SMKT_SMOKE_DETECTOR = 1267

    @classmethod
    def _missing_(cls, value):
        return cls.OTHER


@unique
class EquipmentType(IntEnum):
    """Vivint equipment type."""

    CONTACT = 1
    EMERGENCY = 11
    FREEZE = 6
    MOTION = 2
    TEMPERATURE = 10
    WATER = 8
    UNKNOWN = 0

    @classmethod
    def _missing_(cls, value):
        return cls.UNKNOWN


@unique
class GarageDoorState(IntEnum):
    """Garage door state."""

    UNKNOWN = 0
    CLOSED = 1
    CLOSING = 2
    STOPPED = 3
    OPENING = 4
    OPENED = 5


@unique
class SensorType(IntEnum):
    """Vivint sensor type."""

    AUDIBLE_ALARM = 7
    AUXILIARY_ALARM = 8
    CARBON_MONOXIDE = 14
    DAY_ZONE = 5
    EXIT_ENTRY_1 = 1
    EXIT_ENTRY_2 = 2
    FIRE = 9
    FIRE_WITH_VERIFICATION = 16
    INTERIOR_FOLLOWER = 4
    INTERIOR_WITH_DELAY = 10
    NO_RESPONSE = 23
    PERIMETER = 3
    REPEATER = 25
    SILENT_ALARM = 6
    SILENT_BURGLARY = 24
    UNUSED = 0

    @classmethod
    def _missing_(cls, value):
        return cls.UNUSED



"""Module that implements the AlarmPanel class."""
import logging
from typing import List, Set

from pyvivint.constants import AlarmPanelAttribute as Attributes
from pyvivint.constants import PubNubMessageAttribute as MessageAttributes
from pyvivint.devices import UnknownDevice, VivintDevice, get_device_class
from pyvivint.enums import ArmedState, DeviceType
from pyvivint.utils import first_or_none
from pyvivint.vivintskyapi import VivintSkyApi

_LOGGER = logging.getLogger(__name__)


class AlarmPanel(VivintDevice):
    """Describe a Vivint alarm panel."""

    def __init__(self, data: dict, system: "pyvivint.system.System"):
        super().__init__(data)
        self.system = system
        self.__panel_credentials = None

        # initialize devices
        self.devices: List[VivintDevice] = [
            get_device_class(device_data[Attributes.TYPE])(device_data, self)
            for device_data in self.data[Attributes.DEVICES]
        ]

        # store a reference to the physical panel device
        self.__panel = first_or_none(
            self.devices,
            lambda device: DeviceType(device.data.get(Attributes.TYPE))
            == DeviceType.TOUCH_PANEL,
        )

    @property
    def vivintskyapi(self) -> VivintSkyApi:
        """Instance of VivitSkyApi."""
        return self.system.vivintskyapi

    @property
    def id(self) -> int:
        """Panel's id."""
        return self.data[Attributes.PANEL_ID]

    @property
    def name(self) -> str:
        """Panel's name."""
        return self.system.name

    @property
    def manufacturer(self):
        """Return Vivint as the manufacturer of this panel."""
        return "Vivint"

    @property
    def model(self):
        """Return the model (panel type) of the physical panel."""
        return (
            "Sky Control"
            if self.__panel and self.__panel.data["pant"] == 1
            else "Smart Hub"
        )

    @property
    def software_version(self) -> str:
        """Return the software version of the panel."""
        return self.__panel.software_version if self.__panel else None

    @property
    def partition_id(self) -> int:
        """Panel's partition id."""
        return self.data[Attributes.PARTITION_ID]

    @property
    def is_disarmed(self) -> bool:
        """Return True if alarm is disarmed."""
        return self.get_armed_state() == ArmedState.DISARMED

    @property
    def is_armed_away(self) -> bool:
        """Return True if alarm is in armed away state."""
        return self.get_armed_state() == ArmedState.ARMED_AWAY

    @property
    def is_armed_stay(self) -> bool:
        """Return True if alarm is in armed stay state."""
        return self.get_armed_state() == ArmedState.ARMED_STAY

    def get_armed_state(self):
        """Return the panel's arm state."""
        return self.data[Attributes.STATE]

    async def set_armed_state(self, state: int) -> None:
        """Set the armed state for a panel."""
        _LOGGER.debug(f"setting {self.name} to {ArmedState(state).name}")
        await self.vivintskyapi.set_alarm_state(self.id, self.partition_id, state)

    async def disarm(self) -> None:
        """Disarm the alarm."""
        await self.set_armed_state(ArmedState.DISARMED)

    async def arm_stay(self) -> None:
        """Set the alarm to armed stay."""
        await self.set_armed_state(ArmedState.ARMED_STAY)

    async def arm_away(self) -> None:
        """Set the alarm to armed away."""
        await self.set_armed_state(ArmedState.ARMED_AWAY)

    async def get_panel_credentials(self) -> dict:
        """Gets the panel credentials."""
        if not self.__panel_credentials:
            self.__panel_credentials = await self.vivintskyapi.get_panel_credentials(
                self.id
            )
        return self.__panel_credentials

    def get_devices(
        self, device_types: Set[int] = None, include_unknown_devices: bool = False
    ) -> List[VivintDevice]:
        """Get a list of associated devices."""
        devices = self.devices

        if device_types:
            devices = [
                device
                for device in self.devices
                if device.data[Attributes.TYPE] in device_types
            ]

        if not include_unknown_devices:
            devices = [
                device for device in devices if not isinstance(device, UnknownDevice)
            ]

        return devices

    def refresh(self, data: dict) -> None:
        """Refreshes the alarm panel."""
        self.update_data(data, override=True)

        # update all associated devices
        for device_data in self.data[Attributes.DEVICES]:
            device = first_or_none(
                self.devices, lambda device: device.id == device_data[Attributes.ID]
            )
            if device:
                device.update_data(device_data, override=True)
            else:
                device = get_device_class(device_data[Attributes.TYPE])(
                    device_data, self
                )
                self.devices.append(device)

    def handle_pubnub_message(self, message: dict) -> None:
        """Handles a pubnub message."""
        data = message.get(MessageAttributes.DATA)
        if not data:
            _LOGGER.debug(
                f"ignoring account_partition message for panel {self.id}, partition {self.partition_id} - "
                "no data provided"
            )
            return

        if not data.get(MessageAttributes.DEVICES):
            # this message is for the panel itself
            self.update_data(data)
        else:
            # this is a message for one (or more) of the attached devices
            devices_data = data[MessageAttributes.DEVICES]

            for device_data in devices_data:
                device_id = device_data.get("_id")
                if not device_id:
                    _LOGGER.debug("no device id")
                    continue

                device = first_or_none(
                    self.devices, lambda device: device.id == device_id
                )
                if not device:
                    _LOGGER.debug(
                        f"ignoring message for device {device_id}. Device not found"
                    )
                    continue

                device.handle_pubnub_message(device_data)

                # for the sake of consistency, we need to also update the panel's raw data
                raw_device_data = first_or_none(
                    self.data[Attributes.DEVICES],
                    lambda raw_device_data: raw_device_data["_id"]
                    == device_data["_id"],
                )
                raw_device_data.update(device_data)



#!/usr/bin/env python
"""pyvivint setup script."""

from setuptools import find_packages, setup

install_requires = ["aiohttp>=3.6", "certifi>=2019.9.11"]

setup(
    name="pyvivint",
    version="2021.1.1",
    description="Python module to interact with VivintSky API.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Ovidiu Stateina, Nathan Spencer",
    author_email="ovidiurs@gmail.com",
    license="MIT",
    install_requires=install_requires,
    packages=find_packages(),
    url="https://github.com/ovirs/pyvivint",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Home Automation",
    ],
)
